/* $Id: giqclog.h,v 1.3 2011-09-26 12:54:06 agabasch Exp $
 *
 * This file is part of the GIRAFFE Pipeline
 * Copyright (C) 2002-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2011-09-26 12:54:06 $
 * $Revision: 1.3 $
 * $Name: not supported by cvs2svn $
 */

#ifndef GIQCLOG_H
#define GIQCLOG_H

#include <cpl.h>
/*
#include <cpl_macros.h>
*/

#include <gipaf.h>


#ifdef __cplusplus
extern "C" {
#endif


AmPaf *amber_qclog_open(cxsize id);
cxint amber_qclog_close(AmPaf *qclog);


#ifdef __cplusplus
}
#endif

#endif /* GIQCLOG_H */
